from contextlib import suppress
from traceback import format_exception
from typing import NamedTuple

from kain import Who

from kalib.datastructures import json, serializer
from kalib.text import Str


class Error(NamedTuple):
    type      : str
    arguments : tuple | None
    message   : str | None
    reason    : str
    trace     : tuple | None

    @property
    def as_dict(self):  # typing: ignore[flake8-typing-imports]
        return self._asdict()


@serializer(Error)
def from_error(error):
    return error._asdict()


def exception(e):
    def trim(x):
        return tuple(i.rstrip() for i in x)

    arguments = None
    if (args := getattr(e, 'args', None)):
        result = []
        for item in args:
            with suppress(Exception):
                result.append(str(json.cast(item, throw=True)))
                continue

            with suppress(Exception):
                result.append(Who.Cast(item))
                continue

            with suppress(Exception):
                result.append(repr(item))

        arguments = tuple(result)

    reason = f'{Who(e)}({json.repr(arguments)[1:-1]})'
    if message := getattr(e, 'message', '').strip():
        reason = f'{reason}: {message}'

    return Error(
        Who(e),
        arguments,
        Str(e).strip() or None,
        reason,
        trim(format_exception(e)) or None)



from time import sleep, time

from kain import cache, pin

from kalib import exception
from kalib.dataclass import dataclass

try:
    import redis.exceptions
    from redis import Redis
    from redis.client import PubSub
    from redis.connection import ConnectionPool
    from redis_lock import RedisLock

except ImportError as e:
    raise ImportError(
        'redis & redis_lock is required, install kalib[redis]'
    ) from e


class Pool(dataclass.config):

    Recoverable = (
        ConnectionRefusedError,
        TimeoutError,
        redis.exceptions.BusyLoadingError,
        redis.exceptions.ClusterDownError,
        redis.exceptions.ConnectionError,
        redis.exceptions.TimeoutError,
        redis.exceptions.TryAgainError)

    class PoolConfig(dataclass):
        host                     : str = 'localhost'
        port                     : int = 6379
        db                       : int = 0
        password                 : str | None = None
        socket_timeout           : float | int = 5
        socket_connect_timeout   : float | int = 5
        socket_keepalive         : bool = True
        socket_keepalive_options : dict | None = None
        max_connections          : int = 50
        retry_on_timeout         : bool = True
        health_check_interval    : int = 30
        encoding                 : str = 'utf-8'
        encoding_errors          : str = 'strict'
        decode_responses         : int = True

    @pin.cls.here
    def DefaultPool(cls):  # noqa: N802
        return Redis(connection_pool=ConnectionPool(**cls.PoolConfig.Defaults))

    def __init__(
        self,
        /,
        client  = None,
        poll    = 1.0,
        robust  = True,
        signal  = None,
    ):
        self._client  = client
        self._poll    = poll
        self._signal  = signal
        self._robust  = robust

    @pin
    def pool(self):
        return self._client or self.DefaultPool

    @pin
    def condition(self):
        return self._signal or (lambda: True)

    @property
    def on(self):
        return self.condition()

    #

    @property
    def worker(self):
        client = self.pool
        while self._robust:
            try:
                if client.ping():
                    break

            except self.Recoverable as e:
                self.log.warning(exception(e).reason)

            sleep(self._poll)
        return client

    def __enter__(self):
        return self.worker

    def __exit__(self, *_, **__): ...


class Event(Pool):

    def __init__(
        self,
        name,
        /,
        blocked = True,
        timeout = None,
        ttl     = None,
        **kw,
    ):
        super().__init__(**kw)
        self.name = name
        self._ttl     = ttl
        self._timeout = timeout
        self._value   = self.integer if blocked else -1

    #

    def up(self, ttl: int = 0) -> int:
        with self as cli:
            result = cli.incr(self.name)
            if ttl := int(self._ttl or ttl):
                cli.expire(self.name, ttl)
        return result

    def drop(self) -> int:
        with self as cli:
            return cli.delete(self.name)

    @property
    def value(self) -> int:
        with self as cli:
            return cli.get(self.name)

    @property
    def integer(self) -> int:
        return int(self.value or 0)

    #

    @property
    def updated(self):
        return self._value != self.integer

    def down(self):
        self._value = self.integer

    #

    def changed(
        self,
        timeout  : float | None = None, /,
        poll     : float | None = None,
        infinite : bool = True,
    ) -> bool:
        counter = 0
        start = time()
        wait = float(self._poll or poll)

        if timeout := (timeout or self._timeout):
            deadline = start + timeout

        condition = self.on
        while condition:
            value = self.integer

            if not value:
                self._value = value

            elif self._value != value:
                delta = time() - start

                if counter:
                    self.log.debug(
                        f'{self.name}: {self._value} -> {value} '
                        f'({delta:0.2f}s)')
                self._value = value
                return True

            if timeout:
                wait = min(deadline - time(), self._poll)
                if wait < 0:
                    return infinite

            sleep(wait)
            counter += 1

    def non_zero(self) -> bool:
        counter = 0
        start = time()
        wait = self._poll

        while self.on:
            if self.integer:
                delta = time() - start
                if counter:
                    self.log.debug(f'{self.name} ({delta:0.2f}s)')
                return True

            sleep(wait)
            counter += 1

    __call__ = changed
    __bool__ = non_zero


@cache
def Flag(*args, **kw):  # noqa: N802
    return Event(*args, **kw)


class Lock(RedisLock):

    def __init__(
        self,
        connector,
        name, /,
        timeout = None,
        signal  = None,
    ):
        self.name = name
        self.client = connector
        self._signal = signal
        self._timeout = timeout or 86400 * 365 * 10
        super().__init__(connector, name, blocking_timeout=self._timeout)

    @pin
    def condition(self):
        return self._signal or (lambda: 1)

    #

    def _try_acquire(self) -> bool:
        return self._client.set(self.name, self.token, nx=True, ex=self._ex)

    def _wait_for_message(self, pubsub: PubSub, timeout: int) -> bool:
        deadline = time() + timeout
        condition = self.condition
        while condition():

            message = pubsub.get_message(
                ignore_subscribe_messages=True, timeout=timeout)

            if not message and deadline < time():
                return False

            elif (
                message
                and message['type'] == 'message'
                and message['data'] == self.unlock_message
            ):
                return True

    def acquire(self) -> bool:
        timeout = self._blocking_timeout
        if self._try_acquire():
            return True

        condition = self.condition
        with self._client.pubsub() as pubsub:

            self._subscribe_channel(pubsub)
            deadline = time() + timeout

            while condition():
                self._wait_for_message(pubsub, timeout=timeout)

                if deadline < time():
                    return False

                elif self._try_acquire():
                    return True



from contextlib import suppress
from os import walk
from pathlib import Path
from re import match

from kain import Who, cache, optional, pin, required, sort
from kain.internals import iter_stack

from kalib.datastructures import json
from kalib.loggers import Logging
from kalib.misc import is_python_runtime

DEFAULT_VERSION_RE = r'^(v\d+\.py)$'

logger = Logging.Default


class Git(Logging.Mixin):

    @pin.cls
    def Repo(cls):  # noqa: N802
        return required('git.Repo')

    @pin.cls.here
    def root(cls):
        expect = suppress(required('git.exc.InvalidGitRepositoryError', quiet=True))
        with expect, cls.Repo(
            Path.cwd().resolve(),
            search_parent_directories=True,
        ) as repo:
            return Path(repo.working_tree_dir).resolve()
        cls.log.verbose('not detected')

    @pin.cls.here
    def repo(cls):
        if (path := cls.root):
            return cls.Repo(path, search_parent_directories=True)
        cls.log.verbose('not found')

    @pin.cls.here
    def from_path(cls):
        root = cls.root

        if not root:
            cls.log.warning("feature disabled, git root isn't detected")
            return (lambda *args: None)  # noqa: ARG005

        @cache
        def resolver(path):
            for commit in cls.repo.iter_commits(paths=path, max_count=1):
                return commit
            cls.log.warning(f'{path=} not in repository {root!s} index')

        @cache
        def explorer(path):
            file = Path(path).resolve()

            if not file.is_file():
                cls.log.warning(f'{path=} ({file!s}) not found')
                return

            elif not str(file).startswith(str(root)):
                cls.log.warning(f'{path=} ({file!s}) not in {root!s}')
                return

            return resolver(str(file)[len(str(root)) + 1:])

        return explorer

    @pin.cls.here
    def files(cls):
        result = []
        prefix = len(str(cls.root)) + 1
        for rt, _, fs in walk(cls.root):
            root = Path(rt)
            for f in fs:
                if f.endswith('.py'):
                    result.append(str(root / f)[prefix:])

        return tuple(sort(result))

    @pin.cls.here
    def tree(cls):
        return {
            path: str(cls.from_path(path)) for path in cls.files
            if cls.from_path(path) is not None}

    @pin.cls.here
    def tag(cls):
        with suppress(Exception), cls.repo as git:
            head = git.head.commit

            def get_distance(tag):
                return int(cli.rev_list(
                    '--count', f'{tag.commit.hexsha}..{head.hexsha}'))

            if git.tags:
                cli = required('git.Git')()
                distances = {
                    tag: get_distance(tag)
                    for tag in sort(git.tags, key=lambda x: x.name, reverse=True)}

                if distances:
                    return min(distances, key=distances.get)

    @pin.cls.here
    def version(cls):
        with suppress(ImportError):
            if cls.tag and (version := optional('versioneer.get_version')):
                return version()


@cache
def latest(name, path, pattern=None):
    files = tuple(filter(
        lambda x: match(pattern or DEFAULT_VERSION_RE, x.name),
        Path(path).parent.glob('*.py')))

    if files:
        version = Path(sort(files)[-1]).stem
        module = getattr(__import__(name, globals(), locals(), [version]), version)
        logger.verbose(f'{name} {module=}')
        return module


def add_versioned_path(pattern=None):
    """Add the latest versioned module to the caller's namespace."""

    if not is_python_runtime():
        raise NotImplementedError('this function is for python runtime only')

    def get_call_frame():
        """Get the frame of the caller of the function that called this
        function."""
        for frame in iter_stack(offset=2):
            if frame.filename == __file__:  # skip self
                continue
            if frame.code_context is not None:
                return frame
        raise IndexError

    try:
        pointer = get_call_frame()

    except IndexError:
        raise RuntimeError('could not find the caller frame') from None

    local  = pointer.frame.f_locals
    module = latest(local['__name__'], pointer.filename, pattern)

    to_import = getattr(module, '__all__', ())
    if isinstance(to_import, str):
        # we can override and use as __all__ == 'EntityName'
        to_import = (to_import,)

    local['latest'] = module
    for something in to_import:
        try:
            local[something] = getattr(module, something)

        except AttributeError:

            msg = (
                f"couldn't import {something!a} from module {Who.Name(module)!a}; "
                f"may be it's not found, but declared in {Who.Name(module)}.__all__ = "
                f'{json.repr(to_import)} in {Who.File(module)!r}')

            logger.critical(msg)
            raise ImportError(msg) from None

    return module
